import gc
import abc
import random
import multiprocessing

import numpy as np
import torch as th

from typing import Union, List, Tuple, Any, Dict

from naeural_core.local_libraries.nn.th.training.data.base import BaseDataLoaderFactory, IterableDataset
from naeural_core.local_libraries.nn.th.training_utils import to_device, read_image


class SiameseVisionIterableDataset(IterableDataset):
  def __init__(self, observations : List,
               is_train_dataset : bool,
               are_observations_loaded : bool,
               anchor_label : str,
               load_x_and_y_callback,
               load_x_callback,

               **kwargs):

    super(SiameseVisionIterableDataset, self).__init__(
      observations=observations,
      is_train_dataset=is_train_dataset,
      are_observations_loaded=are_observations_loaded,
      load_x_and_y_callback=load_x_and_y_callback,
      **kwargs)

    self.anchor_label = anchor_label
    self.load_x_callback = load_x_callback
    self.labels_not_counted = []
    return

  def get_data_point(self, idx):
    """
    Returns the data point given by the index parameter.

    If there are multiple anchors for a given index, the anchors are chosen at random.

    For triplet loss, if there is no close image, we will use the anchor again.

    :param idx: datapoint index
    :return:
      if the data class is SiameseMode, ContrastiveLoss or is a dev/test TripletLoss dataset:
        returns a tuple of (anchor, image), the label of the training example and the id of the data point
      else if it is a train dataset of the TripletLoss data class
        returns a tuple of (anchor, close_image, distant image), the label of the training example and the id of the data point
    """
    # First we need to find the location where the given index resides. By computing the cumsum of each location lenght we can find
    #    where our datapoint is found.
    #  For example if the cumsum of location lengths is [45, 200, 350] it means that in the first location we have the
    #    indexes 0 to 44, in the second one we have indexes 45 to 199 and in the last one we have 200 to 350 so we need
    #    to find the first location with the cumsum higher than our given index

    #    argmin([45, 200, 350] < 210) = argmin([True, True, False]) = 2
    #    Explanation: argmin returns the smallest index of the smallest value <=> the index of the first False
    location_idx = np.argmin(self.cumsum_images <= idx)
    location = self._observations[location_idx]

    # The next step is to change the scope of the index, from dataset-scoped to location scoped
    #   (the x-th dataset point is equivalent to the y-th dataset point of the k-th location). We need to find y

    # In order to do so we need to first get the number of points in this location. To do so we subtract the cumsum of the
    #   previous location from the cumsum of the current location
    location_points = self.cumsum_images[location_idx] - (
      self.cumsum_images[location_idx - 1] if location_idx > 0 else 0)
    # Then we find the distance from the global index to the last index of the current location
    location_index_dist = (self.cumsum_images[location_idx] - idx)
    # Finally the location scoped index is given by the difference of the total points minus the distance given by the index
    #   and the final point
    location_scoped_idx = location_points - location_index_dist

    # Next we need to apply the same principle to find the label of the datapoint.
    # First we need to get the counted labels
    filtered_labels = [x for x in location.keys() if x not in self.labels_not_counted]

    # Next, same as before we find the first location with de cumsum higher than the location scoped index
    label_idx = np.argmin(self.cumsum_locations[location_idx] <= location_scoped_idx)

    # Finally, we can get the labeled scoped index (counting from the end) by simply subtracting the cumsum of the label
    #   from the index;
    # NOTE: This will yield a negative value and it's absolute value will represent the index counting from the end
    labeled_scoped_idx = location_scoped_idx - self.cumsum_locations[location_idx][label_idx]

    # Get the index image
    im = location[filtered_labels[label_idx]][labeled_scoped_idx][0]

    # Get the image label
    lbl = self._process_label(location[filtered_labels[label_idx]][labeled_scoped_idx][1])

    ret = self._get_return_tuple(
      im=im,
      lbl=lbl,
      idx=idx,
      location=location
    )

    if not self._are_observations_loaded:
      x, y, idx = ret
      ret = ([self.load_x_callback((im, None))[0] for im in x], y, idx)

    return ret

  def _process_label(self, label):
    return label

  def _get_return_tuple(self, im, lbl, idx, location):
    raise NotImplementedError()

  def __iter__(self):
    self._shuffle()
    worker_indexes = self._get_worker_indexes()

    for i in worker_indexes:
      idx = self._indexes[i]

      yield tuple(self.get_data_point(idx))
    #endfor


class SiameseDataLoaderFactory(BaseDataLoaderFactory):
  def __init__(self, image_height: int, image_width: int, **kwargs):
    self._image_height = image_height
    self._image_width = image_width
    super(SiameseDataLoaderFactory, self).__init__(**kwargs)

  @abc.abstractmethod
  def _get_dataset_class_ref(self):
    raise NotImplementedError

  @abc.abstractmethod
  def _get_not_loaded_observations(self) -> List[Dict[str, List[np.ndarray]]]:
    """
    Method used to load all the image paths in a standardized structure.

    The return structure should be a list of dictionaries. Each list element is a dictionary that represents a location.

    Each location dictionary should have as keys the labels and as values a list of paths to the images from the said label

    Eg.
    [
      {
        'Anchor': ['path/to/location_1/anchor/1.png', 'path/to/location_1/anchor/2.png', 'path/to/location_1/anchor/3.png', ...],
        'label_1': ['path/to/location_1/label_1/1.png', 'path/to/location_1/label_1/2.png', 'path/to/location_1/label_1/3.png', ...],
        ...
      },
      {
        'Anchor': ['path/to/location_2/anchor/1.png', 'path/to/location_2/anchor/2.png',  ...],
        'label_1': ['path/to/location_2/label_1/1.png', 'path/to/location_2/label_1/2.png', ...],
        ...
      },
      ...
    ]
    """
    raise NotImplementedError

  def _load_x(self, observation) -> Tuple[Union[th.tensor, np.ndarray]]:
    path, label = observation
    np_img = read_image(path)
    if label is not None:
      return np_img, label
    return np_img, None

  @to_device('load_device')
  def __to_device(self, *args, **kwargs):
    """
    Overwritten method in order to accept any number of parameters. Needed because we have a variable number of images
      per datapoint.
    """
    params = [arg for arg in args if arg is not None]
    if len(params) == 1:
      return params[0]
    return params

  def _preprocess_on_load(self, x, y):
    x = self._preprocess(x=x, y=y, transform=self._on_load_preprocess)
    return x

  def _preprocess(self, x, y, transform):
    if type(x) in [list, tuple]:
      x = [transform(inp) for inp in x]
    else:
      x = transform(x)

    return x, y

  @to_device('load_device')
  def __to_device(self, *args, **kwargs):
    """
    Overwritten method in order to accept any number of parameters. Needed because we have a variable number of images
      per datapoint.
    """
    params = [arg for arg in args if arg is not None]
    if len(params) == 1:
      return params[0]
    return params


  def load_x_callback(self, observation) -> Tuple[th.tensor]:
    """
    Overwritten method because we don't load the label from here

    """
    self.log.start_timer('one_image', section='{}_preload'.format(self.__name__))

    self.log.start_timer('load_x_and_y', section='{}_preload'.format(self.__name__))
    x = self._load_x(observation)
    self.log.end_timer('load_x_and_y', section='{}_preload'.format(self.__name__))

    self.log.start_timer('to_device', section='{}_preload'.format(self.__name__))
    x = self.__to_device(x)
    self.log.end_timer('to_device', section='{}_preload'.format(self.__name__))

    self.log.start_timer('preprocess_on_load', section='{}_preload'.format(self.__name__))
    x = self._preprocess_on_load(*x)
    self.log.end_timer('preprocess_on_load', section='{}_preload'.format(self.__name__))

    self.log.end_timer('one_image', section='{}_preload'.format(self.__name__))
    return x

  def _get_iterable_dataset_params(self) -> dict:
    """
    Overwritten method because we don't need load the labels separately. The labels are build on __iter__ time in the
      dataset object based on the relation between images / anchors

    The data loaded must be the same as specified in the '_get_not_loaded_observations' method

    Eg:
    [
      {
        'Anchor': [anchor_1, anchor_2, ...],
        'label_1': [label_1_image_1, label_1_image_2, ...],
        ...
      },
      {
        'Anchor': [anchor_1, anchor_2, ...],
        'label_1': [label_1_image_1, label_1_image_2, ...],
        ...
      },
      ...
    ]

    """
    observations_not_loaded = self._get_not_loaded_observations()
    nr_observations = sum([sum([len(x) for x in location.values()]) for location in observations_not_loaded])

    observations = []
    if self._load_observations:
      self.P("Loading the whole dataset ({} observations) to device={} ...".format(len(observations_not_loaded),
                                                                                   self.load_device))
      idx = 0
      for location in observations_not_loaded:
        dct_location = {}
        for cls, lst_paths in location.items():
          if cls not in dct_location:
            dct_location[cls] = []

          for path in lst_paths:
            dct_location[cls].append(self.load_x_callback(path))
            idx += 1

            if self.log.is_main_thread:
              print(
                '\r Observations load progress: {}/{} ({:.2f}%)'.format(idx, nr_observations, 100 * idx / nr_observations),
                flush=True, end=''
              )
            # endif
          #endfor path
        #endfor location
        observations.append(dct_location)
      # endfor

    else:
      observations = observations_not_loaded
    # endif

    dct = {
      'class_ref': self._get_dataset_class_ref(),
      'observations': observations,
      'labels':[],
      'other_params': {
        'load_x_callback': self.load_x_callback,
        **self.data_loader_kwargs
      }
    }
    return dct

  def _get_not_loaded_observations_and_labels(self):
    raise ValueError("Siamese data loaders should implement `_get_not_loaded_observations` method")

  def _load_x_and_y(self, observations, labels, idx):
    raise ValueError("Siamese data loaders should implement `_get_x` method")



"""
IMPORTANT:
  The commands that the `CommunicationManager` responds to are defined directly in the Orchestrator `cmd_handlers_...` methods defined in `core/main/command_handlers.py`
"""

import json
import os
from time import time

import numpy as np

from naeural_core import Logger
from naeural_core import constants as ct
from naeural_core.bc import DefaultBlockEngine, VerifyMessage
from naeural_core.local_libraries import _ConfigHandlerMixin
from naeural_core.manager import Manager


class CommunicationManager(Manager, _ConfigHandlerMixin):
  def __init__(self,
               log: Logger,
               config,
               shmem,
               device_version,
               environment_variables=None,
               avail_commands=[],
               **kwargs):
    self.shmem = shmem
    # TODO: add Shared Memory Manager instance to each plugin under ct.SHMEM.COMM
    self.config = config
    self._device_version = device_version

    self._environment_variables = environment_variables
    self._dct_comm_plugins = None
    self._last_print_info = time()
    self._command_queues = None
    self.__lst_commands_from_self = []
    self._predefined_commands = avail_commands
    self.default_comm_last_active = None

    self.avg_comm_loop_timings = 0

    self.__local_communication_enabled = False # controls if we should start the local communication
    self.__local_communication_active = False # flag that tells us if the local communication has been started

    super(CommunicationManager, self).__init__(log=log, prefix_log='[COMM]', **kwargs)
    return

  def startup(self):
    super().startup()
    self.config_data = self.config
    self._dct_comm_plugins = self._dct_subalterns
    self._command_queues = self._empty_commands()
    return
  
  @property
  def should_bypass_commands_to_self(self):
    return self._environment_variables.get("LOCAL_COMMAND_BYPASS", True)
  
  @property
  def save_received_commands(self):
    return self._environment_variables.get("SAVE_RECEIVED_COMMANDS", False)
  
  @property
  def blockchain_manager(self) -> DefaultBlockEngine:
    return self.shmem[ct.BLOCKCHAIN_MANAGER]
  
  @property
  def is_secured(self):
    return self.log.config_data.get(ct.CONFIG_STARTUP_v2.K_SECURED, False)

  @property
  def _device_id(self):
    _id = self.log.config_data.get(ct.CONFIG_STARTUP_v2.K_EE_ID, '')[:ct.EE_ALIAS_MAX_SIZE]
    return _id  


  @property
  def has_failed_comms(self):
    for comm in self._dct_comm_plugins.values():
      if comm.comm_failed_after_retries:
        self.P("Detected total communication failure on comm {}. This may generate shutdown/restart.".format(comm.__class__.__name__), color='error')
        return True
    return False  


  def _verify_command_signature(self, cmd, verify_allowed=False) -> VerifyMessage:
    result = None
    if self.blockchain_manager is not None:
      result = self.blockchain_manager.verify(cmd, return_full_info=True, verify_allowed=verify_allowed)
      if not result.valid:
        self.P("INCOMING: Command received from sender (verify allowed: {}) addr <{}> signature verification failed: {}".format(
          verify_allowed, result.sender, result.message
          ),color='r'
        )
      else:
        self.P("INCOMING: Command received from sender (verify allowed: {}) addr <{}> signature verification OK.".format(
          verify_allowed, result.sender
        ))
    else:
      raise ValueError("Blockchain Manager is unavailable for verifying the incoming command data")
    return result

  def _load_whitelist_commands(self):
    whitelist_commands_path = os.path.join(self.log.base_folder, ct.WHITELIST_COMMANDS_FILE)
    whitelist_commands = self.log.load_json(whitelist_commands_path)

    if whitelist_commands is None:
      self.P("Creating default whitelist commands file.", verbosity=1)
      self.log.save_json(ct.TEMPLATE_WHITELIST_COMMANDS, whitelist_commands_path)
      whitelist_commands = ct.TEMPLATE_WHITELIST_COMMANDS
    # endif whitelist_commands is None

    return whitelist_commands

  def _reset_whitelist_commands_to_template(self):
    whitelist_commands_path = os.path.join(self.log.base_folder, ct.WHITELIST_COMMANDS_FILE)
    self.log.save_json(ct.TEMPLATE_WHITELIST_COMMANDS, whitelist_commands_path)
    return

  def _verify_whitelist_command(self, cmd):
    template_whitelist_command: list[dict] = self._load_whitelist_commands()
    return any([self.log.match_template(cmd, template) for template in template_whitelist_command])

  def _verify_command_allowed(self, cmd):
    is_authorized = self.blockchain_manager.is_allowed(cmd.get(ct.COMMS.COMM_RECV_MESSAGE.K_SENDER_ADDR, None))
    is_whitelisted = self._verify_whitelist_command(cmd)

    result = is_authorized or is_whitelisted
    result_message = ""

    if is_authorized:
      result_message = "Sender is authorized."
    elif is_whitelisted:
      result_message = "Sender is not authorized but command is allowed due to WHITELIST."
    else:
      result_message = "Sender is not authorized."

    return result, result_message

  def _empty_commands(self):
    return {
      x: [] for x in self._predefined_commands
    }

  def get_received_commands(self):
    # When someone from outside wants to get the received commands, they are automatically refreshed
    # in order to 'make space' for new commands

    # TODO: refactor with deque - this implementation is bad
    ret_received_commands = []
    for k, v in self._command_queues.items(): # for each command type
      for c in v: # for each command of that type
        # each command should be a tuple `payload, sender, session`
        ret_received_commands.append((k, c))
      # endfor each command of that type
    # endfor each command type
    self._command_queues = self._empty_commands()
    return ret_received_commands

  def _get_plugin_class(self, name):
    _module_name, _class_name, _class_def, _class_config = self._get_module_name_and_class(
      locations=ct.PLUGIN_SEARCH.LOC_COMM_PLUGINS,
      name=name,
      suffix=ct.PLUGIN_SEARCH.SUFFIX_COMM_PLUGINS,
      safety_check=True,  # perform safety check
      safe_locations=ct.PLUGIN_SEARCH.SAFE_LOC_COMM_PLUGINS,
    )

    if _class_def is None:
      msg = "Error loading communication plugin '{}'".format(name)
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        info="No code/script defined for communication plugin '{}' in {}".format(
          name, ct.PLUGIN_SEARCH.LOC_COMM_PLUGINS)
      )
      raise ValueError(msg)
    # endif

    return _class_def, _class_config

  def __start_communication(self, config_instance, is_local=False):
    plugin_name = self.config["TYPE"]
    config_instance[ct.EE_ID] = self._device_id
    config_instance[ct.EE_ADDR] = self.blockchain_manager.address

    _class_def, _default_config = self._get_plugin_class(plugin_name)
    id_comm = 0
    for comm_type, paths in self.config["INSTANCES"].items():
      id_comm += 1

      if not is_local:
        # the default connection, used to communicate with the network
        comm_type = comm_type.upper()
      else:
        # the local connection, used for local communication
        comm_type = "L_" + comm_type.upper()
      send_channel_name = paths.get("SEND_TO", None)
      recv_channel_name = paths.get("RECV_FROM", None)
      if send_channel_name is not None:
        send_channel_name = send_channel_name.upper()
      if recv_channel_name is not None:
        recv_channel_name = recv_channel_name.upper()

      try:
        comm_plugin = _class_def(
          log=self.log,
          shmem=self.shmem,
          signature=plugin_name,
          comm_type=comm_type,
          default_config=_default_config,
          upstream_config=config_instance,
          environment_variables=self._environment_variables,
          send_channel_name=send_channel_name,
          recv_channel_name=recv_channel_name,
          timers_section=str(id_comm) + '_' + comm_type,
        )
        comm_plugin.validate(raise_if_error=True)
        comm_plugin.start()
        self.add_subaltern(comm_type, comm_plugin)
      except Exception as exc:
        msg = "Exception '{}' when initializing communication plugin {}".format(exc, plugin_name)
        self.P(msg, color='r')
        self._create_notification(
          notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
          msg=msg,
          autocomplete_info=True
        )
        raise exc
      # end try-except
    # endfor
    return

  def __maybe_start_local_communication(self):
    if self.__local_communication_active:
      self.P("Local communication has already been started", color='y')
      return

    if not self.__local_communication_enabled:
      self.P("Local communication is disabled", color='y')
      return

    # TODO: maybe change LOCAL_PARAMS to something else
    if "LOCAL_PARAMS" not in self.config:
      self.config["LOCAL_PARAMS"] = {
        "HOST": "localhost",
        "PASS": "",
        "PORT": 1883,
        "USER": "",
        "QOS": 0,
        "SECURED": 0,
      }
      self.P("Local communication is not configured! Using default values: localhost:1883 no user and no password")

    local_config_instance = {**self.config["PARAMS"], **self.config["LOCAL_PARAMS"]} 
    self.__start_communication(local_config_instance, is_local=True)
    self.__local_communication_active = True
    return

  def start_communication(self):
    self.__start_communication(self.config["PARAMS"])

    # we use the local communication server for development and on-edge deployments
    self.__maybe_start_local_communication()
    return

  def __select_communicators_and_send(self, data, communicator_name, local_only=False):
    if isinstance(data, dict):
      local_only = local_only or data.get('USE_LOCAL_COMMS_ONLY', False)

    central_communicator = self._dct_comm_plugins[communicator_name]
    local_communicator = self._dct_comm_plugins.get("L_" + communicator_name)

    # maybe the local communicator is disabled
    # TODO: maybe remove redundant check  (local_communicator is not None)
    if self.__local_communication_enabled and local_communicator is not None and self.__local_communication_active:
          local_communicator.send(data)

    if not local_only:
      central_communicator.send(data)

    return central_communicator

  def send(self, data, event_type='PAYLOAD'):
    if event_type == 'COMMAND':
      receiver_id, receiver_addr, local_only, command_type, command_content = data
      
      # in some cases the receiver_addrs is a non-prefixed address
      receiver_addr = self.blockchain_manager.maybe_add_prefix(receiver_addr) # add prefix if needed

      is_command_to_self = receiver_addr == self.blockchain_manager.address

      if self.should_bypass_commands_to_self and is_command_to_self:
        # if we are the receiver, we bypass sending and receiving the command to the network
        self.P("Bypassing network communication for {} local command to self".format(command_type), color='y')
        self.__lst_commands_from_self.append((command_type, command_content))
      else:
        # we send the command to the network
        # EE_ID is the receiver ID for command and this is a ambiguity that we need to resolve
        # as well as EE_ADDR is the receiver address
        # but in normal payload messages, EE_ADDR is the sender address as EE_ID is the device ID
        command = {
          ct.EE_ADDR: receiver_addr,
          ct.EE_ID: receiver_id, 
          ct.COMMS.COMM_SEND_MESSAGE.K_ACTION: command_type,
          ct.COMMS.COMM_SEND_MESSAGE.K_PAYLOAD: command_content,
          ct.COMMS.COMM_SEND_MESSAGE.K_INITIATOR_ID: self._device_id,
          ct.COMMS.COMM_SEND_MESSAGE.K_SESSION_ID: self.log.session_id,
          ct.COMMS.COMM_SEND_MESSAGE.K_TIME: self.log.now_str()
        }
        # commands generated by plugins that are in local mode will be sent only locally
        # TODO: check if this is the correct behavior 
        self.__select_communicators_and_send((receiver_id, receiver_addr, command), ct.COMMS.COMMUNICATION_COMMAND_AND_CONTROL, local_only=local_only)
    else:
      message = {
        ct.EE_ADDR: self.blockchain_manager.address,
        ct.EE_ID: self._device_id,
        'EE_EVENT_TYPE': event_type,
        'EE_VERSION': self._device_version,
        **data
      }

      if event_type == 'PAYLOAD':
        # payloads should be sent to both local and central communicators
        # payloads need to be sent to local communicators only if a flag is present (flag checked in __select_communicators_and_send)
        communicator = self.__select_communicators_and_send(message, ct.COMMS.COMMUNICATION_DEFAULT, local_only=False)
        if len(communicator.loop_timings) > 0:
          self.avg_comm_loop_timings = np.mean(communicator.loop_timings)
      elif event_type == 'NOTIFICATION':
        self.__select_communicators_and_send(message, ct.COMMS.COMMUNICATION_NOTIFICATIONS, local_only=False)
      elif event_type == 'HEARTBEAT':
        self.__select_communicators_and_send(message, ct.COMMS.COMMUNICATION_HEARTBEATS, local_only=False)
      # endif
    # endif
    return

  def close(self):
    self.P("Clossing all comm plugins", color='y')
    for comm in self._dct_comm_plugins.values():
      comm.stop()
    self.__local_communication_active = False
    self.P("Done closing all comm plugins.", color='y')
    return
  
  def __maybe_close_local_communication(self):
    self.P("Clossing all local comm plugins", color='y')
    if not self.__local_communication_active:
      self.P("Local communication is already closed", color='y')
      return

    comms_to_close = [comm for comm in self._dct_comm_plugins.keys() if comm.startswith("L_")]
    for comm_name in comms_to_close:
      comm = self._dct_comm_plugins.pop(comm_name)
      comm.stop()
      del comm
    # endfor
    self.__local_communication_active = False
    self.P("Done closing all local comm plugins.", color='y')
    return

  def enable_local_communication(self):
    """
    Enable local communication. This method will start the local communication if it is not already started.
    """
    self.__local_communication_enabled = True
    self.__maybe_start_local_communication()
    return

  def disable_local_communication(self):
    """
    Disable local communication. This method will stop the local communication if it is started.
    """
    self.__local_communication_enabled = False
    self.__maybe_close_local_communication()
    return

  def maybe_process_incoming(self):
    communicator_for_config = self._dct_comm_plugins[ct.COMMS.COMMUNICATION_HEARTBEATS]
    local_communicator_for_config = self._dct_comm_plugins.get("L_" + ct.COMMS.COMMUNICATION_HEARTBEATS)
    incoming_commands = communicator_for_config.get_messages()

    if local_communicator_for_config is not None:
      # process local commands
      incoming_commands_local = local_communicator_for_config.get_messages()
      incoming_commands.extend(incoming_commands_local)

    for json_msg in incoming_commands:
      self.process_command_message(json_msg)
    self.process_commands_from_self()
    return self.get_received_commands()

  def get_total_bandwidth(self):
    inkB, outkB = 0, 0
    keys = list(self._dct_comm_plugins.keys())
    for name in keys:
      comm = self._dct_comm_plugins.get(name)
      if comm is not None:
        inkB += comm.get_incoming_bandwidth()
        outkB += comm.get_outgoing_bandwidth()
    return inkB, outkB

  def get_comms_status(self):
    dct_stats = {}
    try:
      self.default_comm_last_active = self.log.time_to_str(
          self._dct_comm_plugins[ct.COMMS.COMMUNICATION_DEFAULT].last_activity_time
      )
    except Exception as exc:
      self.default_comm_last_active = f"ERROR: {exc}"
    # endtry get last activity time
    
    keys = list(self._dct_comm_plugins.keys())
    for name in keys:
      comm = self._dct_comm_plugins.get(name)
      if comm is not None:
        errors, times = comm.get_error_report()
        dct_stats[name] = {
          'SVR': comm.has_server_conn,
          'RCV': comm.has_recv_conn,
          'SND': comm.has_send_conn,
          'ACT': comm.last_activity_time,
          'ADDR': comm.server_address,
          'FAILS': len(errors),
          'ERROR': errors[-1] if len(errors) > 0 else None,
          'ERRTM': times[-1] if len(times) > 0 else None,
          ct.HB.COMM_INFO.IN_KB: comm.get_incoming_bandwidth(),
          ct.HB.COMM_INFO.OUT_KB: comm.get_outgoing_bandwidth(),
        }
    return dct_stats

  def maybe_show_info(self):
    now = time()
    if (now - self._last_print_info) >= ct.COMMS.COMM_SECS_SHOW_INFO:
      communicator = self._dct_comm_plugins[ct.COMMS.COMMUNICATION_DEFAULT]
      dct_stats = self.get_comms_status()
      keys = list(dct_stats.keys())
      ml = max([len(k) for k in keys])
      lines = []
      inkB, outkB = 0, 0
      for n in dct_stats:
        inkB += dct_stats[n][ct.HB.COMM_INFO.IN_KB]
        outkB += dct_stats[n][ct.HB.COMM_INFO.OUT_KB]
        active = dct_stats[n]['ACT']
        line = '    {}: live {}, conn:{}, rcv:{}, snd:{}, fails:{}, err: {}, {}'.format(
          n + ' ' * (ml - len(n)),
          self.log.time_to_str(active),
          int(dct_stats[n]['SVR']),
          int(dct_stats[n]['RCV']),
          int(dct_stats[n]['SND']),
          dct_stats[n]['FAILS'],
          dct_stats[n]['ERRTM'],
          dct_stats[n]['ADDR'],
        )
        lines.append(line)

      self.P("Showing comms statistics (In/Out/Total {:.2f} kB / {:.2f} kB / {:.2f} kB):\n{}".format(
        inkB, outkB, inkB + outkB,
        "\n".join(lines)),
        color=ct.COLORS.COMM
      )
      self._last_print_info = now
      statistics_payloads_trip = communicator.statistics_payloads_trip
      if len(statistics_payloads_trip):
        self.P(statistics_payloads_trip)
    # endif
    return

  def _save_input_command(self, payload):
    fn = '{}.json'.format(self.log.now_str())
    self.log.save_output_json(
      data_json=payload,
      fname=fn,
      subfolder_path='received_commands',
      verbose=True
    )
    return

  def process_commands_from_self(self):
    for action, payload in self.__lst_commands_from_self:
      # we populate the fields with our data because the message is supposed to be from us
      self.process_decrypted_command(
        action=action,
        payload=payload,
        sender_addr=self.blockchain_manager.address,
        initiator_id=self._device_id,
        session_id=self.log.session_id,
        validated_command=True,
      )
    self.__lst_commands_from_self = []
    return

  def process_decrypted_command(self, action, payload, sender_addr=None, initiator_id=None, session_id=None, validated_command=False, json_msg=None):
    """
    This method is called to process a command that has been decrypted.
    We assume the command has been validated at this point, so we can proceed with processing it.
    We can discard a command if action is None or if the action is not recognized.

    action: str
      The action to be performed

    payload: dict
      The payload of the command

    sender_addr: str
      The address of the sender

    initiator_id: str
      The ID of the initiator

    session_id: str
      The ID of the session

    validated_command: bool
      Whether the command has been validated or not

    json_msg: dict
      The original JSON message
    """
    if action is not None:
      action = action.upper()
      if payload is None:
        self.P("  Message with action '{}' does not contain payload".format(
          action), color='y'
        )
        payload = {}  # initialize payload
      # endif no payload

      if isinstance(payload, dict):
        # we add the sender address to the payload
        payload[ct.COMMS.COMM_RECV_MESSAGE.K_SENDER_ADDR] = sender_addr
        # we add or modify payload session & initiator for downstream tasks
        if payload.get(ct.COMMS.COMM_RECV_MESSAGE.K_INITIATOR_ID) is None or initiator_id is not None:
          payload[ct.COMMS.COMM_RECV_MESSAGE.K_INITIATOR_ID] = initiator_id
        if payload.get(ct.COMMS.COMM_RECV_MESSAGE.K_SESSION_ID) is None or session_id is not None:
          payload[ct.COMMS.COMM_RECV_MESSAGE.K_SESSION_ID] = session_id
        # we send the message that this command was validated or not
        payload[ct.COMMS.COMM_RECV_MESSAGE.K_VALIDATED] = validated_command
      # endif

      if action not in self._command_queues.keys():
        self.P("  '{}' - command unknown".format(action), color='y')
      else:
        # each command is a tuple as below
        self._command_queues[action].append((payload, sender_addr, initiator_id, session_id))
      # endif
    else:
      self.P('  Message does not contain action. Nothing to process...', color='y')
      self.P('  Message received: \n{}'.format(json_msg), color='y')
    return

  def process_command_message(self, json_msg):
    device_id = json_msg.get(ct.EE_ID, None)
    action = json_msg.get(ct.COMMS.COMM_RECV_MESSAGE.K_ACTION, None)
    payload = json_msg.get(ct.COMMS.COMM_RECV_MESSAGE.K_PAYLOAD, None)
    initiator_id = json_msg.get(ct.COMMS.COMM_RECV_MESSAGE.K_INITIATOR_ID, None)
    session_id = json_msg.get(ct.COMMS.COMM_RECV_MESSAGE.K_SESSION_ID, None)
    sender_addr = json_msg.get(ct.COMMS.COMM_RECV_MESSAGE.K_SENDER_ADDR, None)
    # next line is actually redundant as it should always be the same as the device_id
    # it was left here for compatibiliy with the normal payload structure
    dest_addr = json_msg.get(ct.PAYLOAD_DATA.EE_DESTINATION, None)
    
    if self.save_received_commands:
      self.P("INCOMING: Saving received command ...")
      self._save_input_command(json_msg)
    #endif save_received_commands

    self.P("INCOMING: Received message with action '{}' from <{}:{}>".format(
      action, initiator_id, session_id),
      color='y'
    )
    failed = False

    is_eth_address = self.blockchain_manager.eth_enabled and (self.blockchain_manager.eth_address == device_id)

    if device_id != self._device_id and device_id != self.blockchain_manager.address and not is_eth_address:
      self.P('INCOMING:   Message is not for the current device {} != {} ({})'.format(
        device_id, self._device_id, self.blockchain_manager.address), color='y')
      failed = True
      
    ### signature verification    
    # allowed list will be checked later due to the fact that we will do one shot allowed check and 
    # command-whitelist check conditioned by the decryption of the command itsels (to check for potential whitelist)
    _verify_allowed_at_signature = False
    verify_msg = self._verify_command_signature(json_msg, verify_allowed=_verify_allowed_at_signature) 
    if not verify_msg.valid:
      if self.is_secured:
        msg = "INCOMING: Received invalid command from {}({}):{} due to '{}'. Command will be DROPPED.".format(
          initiator_id, verify_msg.sender, json_msg, verify_msg.message
        )
        failed = True
      else:
        msg = "INCOMING: Received invalid command from {}({}):{} due to '{}'. Command is accepted due to UNSECURED node.".format(
          initiator_id, verify_msg.sender, json_msg, verify_msg.message
        )
        failed = False
      #endif failed or not
      notif_code=ct.NOTIFICATION_CODES.COMM_RECEIVED_BAD_COMMAND if failed else None
      # TODO: in future maybe we should NOT send notifications for invalid received commands
      #       just ignore them and ze zee zeet
      self.P(msg, color='error')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING,
        msg=msg,
        session_id=session_id,
        notif_code=notif_code,
        displayed=False,
      )
    else:
      if _verify_allowed_at_signature:
        msg = "INCOMING: Command from {}({}) signature VALIDATED (verify allowed: {}).".format(
          initiator_id, verify_msg.sender, _verify_allowed_at_signature
        )
        self.P(msg)
    ### end signature verification
    validated_command = verify_msg.valid 


    if failed:
      self.P("INCOMING:   Message dropped.", color='r')      
    else:
      # Not failed, lets process
      is_encrypted = json_msg.get(ct.COMMS.COMM_RECV_MESSAGE.K_EE_IS_ENCRYPTED, False)
      if is_encrypted:
        encrypted_data = json_msg.pop(ct.COMMS.COMM_RECV_MESSAGE.K_EE_ENCRYPTED_DATA, None)
        # TODO: reduce log redundancy
        if dest_addr is not None and dest_addr != self.blockchain_manager.address:
          self.P("INCOMING:   Message is encrypted but not for this device. Decryption will fail.", color='r')
        str_data = self.blockchain_manager.decrypt(
          encrypted_data_b64=encrypted_data, 
          sender_address=sender_addr,
          debug=False, 
        )
        if str_data is None:
          self.P("INCOMING:   Decryption failed with SDKv{}. Message will be probably dropped.".format(
            self.log.version), 
            color='r'
          )
        else:
          try:
            dict_data = json.loads(str_data)
            action = dict_data.get(ct.COMMS.COMM_RECV_MESSAGE.K_ACTION, None)
            payload = dict_data.get(ct.COMMS.COMM_RECV_MESSAGE.K_PAYLOAD, None)
            json_msg.update(dict_data)
          except Exception as e:
            self.P("INCOMING: Error while decrypting message: {}\n{}".format(str_data, e), color='r')
      # endif is_encrypted

      # TODO: change this value to False when all participants send encrypted messages
      if not is_encrypted and not self._environment_variables.get("ACCEPT_UNENCRYPTED_COMMANDS", True):
        self.P("INCOMING:   Message is not encrypted. Message dropped because `ACCEPT_UNENCRYPTED_COMMANDS=False`.", color='r')
      else:
        # now that the message is decrypted, we can check if it is allowed or not as well
        # as if it is whitelisted or not
        self.P("INCOMING: Message pre-processed (is_encrypted: {}), verifying allow list...".format(
          is_encrypted
        ))
        allowed, allowed_msg = self._verify_command_allowed(json_msg)
        self.P("INCOMING:   Command allowed check: {}. Reason: {}".format(allowed, allowed_msg), color=None if allowed else 'r')
        if allowed:
          self.process_decrypted_command(
            action=action,
            payload=payload,
            sender_addr=sender_addr,
            initiator_id=initiator_id,
            session_id=session_id,
            validated_command=validated_command,
            json_msg=json_msg
          )
    return

  def validate_macro(self):
    communication = self.config_data
    if communication is None:
      msg = "'COMMUNICATION' is not configured in `config_app.txt`"
      self.add_error(msg)
      return

    plugin_name = communication.get("TYPE", None)
    params = communication.get("PARAMS", None)
    dct_instances = communication.get("INSTANCES", None)

    if plugin_name is None:
      msg = "Parameter 'TYPE' is not configured for 'COMMUNICATION' in `config_app.txt`"
      self.add_error(msg)

    if params is None:
      msg = "Parameter 'PARAMS' is not configured for 'COMMUNICATION' in `config_app.txt`"
      self.add_error(msg)
    else:
      found_channels = []
      for k in params:
        if k in ct.COMMS.COMMUNICATION_VALID_CHANNELS:
          found_channels.append(k)

      if len(set(ct.COMMS.COMMUNICATION_VALID_CHANNELS) - set(found_channels)) != 0:
        self.add_error("Make sure that all communication channels {} are configured as 'PARAMS' for 'COMMUNICATION' in `config_app.txt`".format(
          ct.COMMS.COMMUNICATION_VALID_CHANNELS))
      port = params.get("PORT", None)
      if port is not None:
        if not isinstance(port, int):
          self.add_warning("Parameter 'PORT' is not an integer for 'COMMUNICATION' in `config_app.txt` - casting to int")
          params["PORT"] = int(port)
        # endif not int
      # endif port
          
    #endif params

    if dct_instances is None:
      msg = "Parameter 'INSTANCES' is not configured for 'COMMUNICATION' in `config_app.txt`"
      self.add_error(msg)
    else:
      found_instances = []
      for comm_type, paths in dct_instances.items():
        comm_type = comm_type.upper()
        send_channel_name = paths.get("SEND_TO", None)
        recv_channel_name = paths.get("RECV_FROM", None)
        if send_channel_name is not None:
          send_channel_name = send_channel_name.upper()
        if recv_channel_name is not None:
          recv_channel_name = recv_channel_name.upper()
        if comm_type not in ct.COMMS.COMMUNICATION_VALID_TYPES:
          msg = "Parameter 'INSTANCE' is misconfigured for 'COMMUNICATION' in `config_app.txt` - unknown instance {}; please try one of these: {}".format(
            comm_type, ct.COMMS.COMMUNICATION_VALID_TYPES)
          self.add_error(msg)

        if send_channel_name is not None and send_channel_name not in ct.COMMS.COMMUNICATION_VALID_CHANNELS:
          msg = "Parameter 'INSTANCE' is misconfigured for 'COMMUNICATION' in `config_app.txt` - for instance {} 'SEND_TO' is not valid; please try one of these: {}".format(
            comm_type, ct.COMMS.COMMUNICATION_VALID_CHANNELS)
          self.add_error(msg)

        if recv_channel_name is not None and recv_channel_name not in ct.COMMS.COMMUNICATION_VALID_CHANNELS:
          msg = "Parameter 'INSTANCE' is misconfigured for 'COMMUNICATION' in `config_app.txt` - for instance {} 'RECV_FROM' is not valid; please try one of these: {}".format(
            comm_type, ct.COMMS.COMMUNICATION_VALID_CHANNELS)
          self.add_error(msg)

        found_instances.append(comm_type)
      # endfor
      if len(set(ct.COMMS.COMMUNICATION_VALID_TYPES) - set(found_instances)) != 0:
        self.add_error("Make sure that all communication instances {} are configured as 'INSTANCES' for 'COMMUNICATION' in `config_app.txt`".format(
          ct.COMMS.COMMUNICATION_VALID_TYPES))

    # endif

    return



import os
import requests
import json
import traceback

_FOLDER = './personas'

DIRECT_USER = 'direct-user'

class OpenAIApp(object):
  def __init__(self, persona, user=None, log=None, persona_location=None):
    assert isinstance(persona, str), "`Persona` must be a string"
    if persona_location is None:
      persona_location = _FOLDER
    self.__persona_location = persona_location
    self.log = None
    self.data = {}
    self.persona = persona.lower()
    if user is not None:
      user = user.lower()
    self.user = user
    self.load_persona()
    return
  
  def P(self, s, color=None):
    if self.log is None:
      print(s, flush=True)
    else:
      self.log.P(s, color=color)
    return
  
  
  def list_models(self):
    url = 'https://api.openai.com/v1/models'
    res = self.request(data=None, url=url, content_path=['data'], method='GET')    
    return res
    
  
  def _get_empty_data(self, key=None, model=None, user=None, init=None, functions=None, function_call=None):
    if user is None:
      if self.user is None:
        raise ValueError("_get_empty_data without specific USER on a user-less chat engine")
      else:
        user = self.user
    user = user.lower()
      
    if init is None:
      init = self._init
    if key is None:
      key = self._key
    if model is None:
      model = self._model
    if self._is_chatbot:
      data = {
        "model" : self._model,
        "user"  : user,
        "messages" : []
      }
      if init is not None:
        data['messages'].append({"role" : "system", "content" : init})
    else:
      data = {
        "model" : self._model,
      }
    
    if functions is not None:
      data.update({"functions": functions})
    if function_call is not None:
      data.update({"function_call": function_call})
    return data  
  
  
  def _add_message(self, content, user=None, role='user', data=None):
    if data is None:
      if user is None:
        if self.user is None:
          raise ValueError("_add_message without specific USER/data on a user-less chat engine")
        else:
          user = self.user      
      user = user.lower()        
      if user not in self.data:
        self.reset(user=user)
      data = self.data[user]
    if self._is_chatbot:
      turn = {"role" : role, "content" : content}
      data['messages'].append(turn)
    else:
      data = {
        "model" : self._model,
        "prompt": content,
      }      
    data = {
      **self._params,
      **data,
    }
    return data
  
  
  def request(self, data, debug=False, url=None, content_path=None, method='POST'):
    if url is None:
      url = self._url
    if content_path is None:
      content_path = self._content_path
    response = None
    result = None
    headers = None
    try:    
      if method.upper() == 'POST':
        func = requests.post
      else:
        func = requests.get
      headers = {
          "Content-Type": "application/json", 
          "Authorization": "Bearer {}".format(self._key)
      }
      response = func(
        url,
        headers=headers,
        json=data,
      )
      response = response.json()
      result = response
      for path in content_path:
        result = result[path]      
      if debug:
        print("{}: {}".format(self.__class__.__name__, response))
    except Exception as exc:
      err_msg = "Request failed with '{}', reponse: {}, headers: {}".format(exc, response, headers)
      raise ValueError(err_msg)
    return result

    
  def ask(self, question, user=None, debug=False):
    if user is None:
      user = self.user
      if self.user is None:
        raise ValueError("ask-ed a question without specific USER on a user-less chat engine")
    user = user.lower()
        
    self._add_message(question, user=user, role='user')
    data = self.data[user]
    result = self.request(data, debug=debug)
    if result is not None:
      self._add_message(result, user=user, role='assistant')
    return result
  
  
  def ask_direct(self, question, debug=False):
    self.reset(user=DIRECT_USER)
    data = self._add_message(question, user=DIRECT_USER)
    result = self.request(data, debug=False)
    return result


  def load_persona(self, persona=None, user=DIRECT_USER):
    if persona is None:
      persona = self.persona
    found = False
    if ".txt" not in persona:
      fn = os.path.join(self.__persona_location, persona + ".txt")
      found = os.path.isfile(fn)
    if not found and ".json" not in persona:
      fn = os.path.join(self.__persona_location, persona, '.json')
      found = os.path.isfile(fn)
    if not found:
      fn = os.path.join(self.__persona_location, persona)
      found = os.path.isfile(fn)
          
    assert found, "Unknown chatbot persona '{}'".format(persona)
    
    with open(fn, "rt") as fh:
      _data = json.load(fh)
    
    self._key = _data['api_key']
    if self._key is None:
      self._key = os.environ['EE_OPENAI']
    self._init = _data.get('init')
    self._url = _data['url']
    self._model = _data['model']
    self._content_path = _data['content_path']
    self._is_chatbot = _data['is_chat']
    self._params = _data.get('params', {})
    self._functions = _data.get('functions')
    self._function_call = _data.get('function_call')
    self.reset(functions=self._functions, function_call=self._function_call, user=user)
    return


  def reset(self, key=None, user=None, init=None, model=None, functions=None, function_call=None):
    if user is None:
      user = self.user
      if user is None:
        raise ValueError("Reset received without specific USER on a user-less chat engine")

    user = user.lower()
    self.data[user] = self._get_empty_data(
      key=key, 
      model=model, 
      init=init, 
      user=user,
      functions=functions, 
      function_call=function_call,
    )
    return
  
  
  
if __name__ == '__main__':
  DATA_TEST = False
        
  if DATA_TEST:
    query = 'Please generate in json format quizes where the answer is single word and completes the question. Propose another 4 wrong asnwers. Thematic is elementary school math. The questions should be amusing. Simplify the json format such as the "question" is "q", "correct_answer" is "a", "wrong_answers" is "wa" and the language "l":"en" or "l":"ro". Write the json puting each object in one line with the keys in the following order: "l", "q", "a", "wa" and end the line with a comma. Generate a total of 50 quizes for Romanian language'
    eng = OpenAIApp(persona="Worker", user='Andrei')
    res = eng.ask_direct(query)
    print(res)


from copy import deepcopy

class _DataAPIMixin(object):
  """
  Mixin for `BasePluginExecutor` that transparently handles the work with a plugin instance input.

  Most common methods
  -------------------
    * dataapi_image
    * dataapi_image_instance_inferences
    * dataapi_image_plugin_inferences

  Plugin instance input examples
  ------------------------------
  1.
  {
    'STREAM_NAME' : 'simple_image_stream',
    'STREAM_METADATA' : {Dictionary with general stream metadata},
    'INPUTS' : [
      {
        'IMG' : np.ndarray,
        'STRUCT_DATA' : None,
        'INIT_DATA' : None,
        'TYPE' : 'IMG',
        'METADATA' : {Dictionary with current input metadata}
      }
    ],
    'INFERENCES' : {
      'object_detection_model' : [
        [{'TLBR_POS' : ...}, {'TLBR_POS' : ...}] # there are 2 detections on the first (single) image
      ]
    },

    'INFERENCES_META' : {
      'object_detection_model' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : 'IMG'},
    }
  }

  2.
  {
    'STREAM_NAME' : 'multi_modal_2_images_one_sensor_stream',
    'STREAM_METADATA' : {Dictionary with general stream metadata},
    'INPUTS' : [
      {
        'IMG' : np.ndarray(1),
        'STRUCT_DATA' : None,
        'INIT_DATA' : None,
        'TYPE' : 'IMG',
        'METADATA' : {Dictionary with current input metadata}
      },

      {
        'IMG' : None,
        'STRUCT_DATA' : an_object,
        'INIT_DATA' : None,
        'TYPE' : 'STRUCT_DATA',
        'METADATA' : {Dictionary with current input metadata}
      },

      {
        'IMG' : np.ndarray(2),
        'STRUCT_DATA' : None,
        'INIT_DATA' : None,
        'TYPE' : 'IMG',
        'METADATA' : {Dictionary with current input metadata}
      },
    ],
    'INFERENCES' : {
      'object_detection_model' : [ # `object_detection_model` returns a list with 2 elements (because there are 2 images in 'INPUTS')
        [{'TLBR_POS' : ...}, {'TLBR_POS' : ...}, {'TLBR_POS' : ...}], # there are 3 detections on the first image
        [{'TLBR_POS' : ...}] # there is 1 detection on the first image
      ],

      'anomaly_detection_model' : [ # `anomaly_detection_model` returns a list with 1 elemnt (because there is 1 structured data in 'INPUTS')
        'True/False'
      ]
    },
    'INFERENCES_META' : {
      'object_detection_model' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : 'IMG'},
      'anomaly_detection_model' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : 'STRUCT_DATA'}
    }
  }

  Understanding "INPUTS" in-depth
  -------------------------------
  The value for "INPUTS" key in the plugin instance input is a list of dictionaries (depending on how many inputs/substreams a stream collects -
  it may collect one frame from a video feed, or it may collect one frame and one chunk of structured data, without any
  restriction on the number of the inputs).
  A dictionary in the "INPUTS" list - i.e. a substream - has a clear structure:
    {
      "IMG" : None or np.ndarray (None if the input is a structured data, np.ndarray otherwise - mutual exclusive with "STRUCT_DATA"),
      "STRUCT_DATA" : None or object (None if the input is an image, object otherwise - mutual exlusive with "IMG",
      "INIT_DATA" : None or object (some initial data that is provided by the substream - None if the substream does not provide initial data),
      "TYPE" : "IMG" or "STRUCT_DATA" (specifies which field is completed),
      "METADATA" : dict (dictionary with current input metadata)
    }
  """

  def __init__(self):
    super(_DataAPIMixin, self).__init__()
    return

  def dataapi_full_input(self):
    """
    Returns
    -------
    dict
      full input, as it comes from upstream (empty dictionary if there is not data from upstream):
      {
        'STREAM_NAME' : 'multi_modal_2_images_one_sensor_stream',
        'STREAM_METADATA' : ...,
        'INPUTS' : [{...}, {...}, {...}],
        'INFERENCES : {
          'object_detection_model' : [
            [{'TLBR_POS' : ...}, {'TLBR_POS' : ...}, {'TLBR_POS' : ...}],
            [{'TLBR_POS' : ...}]
          ],

          'anomaly_detection_model' : [
            'True/False'
          ]
        },
        'INFERENCES_META' : {
          'object_detection_model' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : 'IMG'},
          'anomaly_detection_model' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : 'STRUCT_DATA'}
        }
      }
    """
    if self.inputs is not None:
      res = self.inputs
    else:
      res = {}
    return res
  
  def dataapi_stream_info(self):
    dct_inp = self.dataapi_full_input()
    return {k:v for k,v in dct_inp.items() if k not in ['INPUTS', 'INFERENCES', 'INFERENCES_META']}

  def dataapi_plugin_input(self):
    """
    Alias for `self.dataapi_full_input`

    Returns
    -------
    dict
      full input, as it comes from upstream (empty dictionary if there is not data from upstream)
    """
    return self.dataapi_full_input()

  def dataapi_received_input(self):
    """
    Returns
    -------
    bool
      whether input received from upstream or not (a plugin can run also without input)
    """
    return bool(self.dataapi_full_input())

  """
  Section for methods that return the first level information of the received input.
  """
  if True:
    def dataapi_stream_name(self):
      """
      Returns
      -------
      str
        the name of the stream that sends data to the current plugin instance
      """
      return self.dataapi_full_input().get('STREAM_NAME', None)    
    
    def dataapi_stream_metadata(self):
      """
      This function serves returns all the params that configured the current execution
      pipeline where the plugin instance is executed.
      
      Returns
      -------
      dict
        the metadata of the stream that sends data to the current plugin instance
      """
      return self.dataapi_full_input().get('STREAM_METADATA', {})

    def dataapi_inputs(self):
      """
      Returns
      -------
      list[dict]
        the inputs of the stream that sends data to the current plugin instance
      """
      return self.dataapi_full_input().get('INPUTS', [])

    def dataapi_inferences(self, squeeze=False):
      """
      Returns
      -------
      dict{str:list}
        the inferences that come from the serving plugins configured for the current plugin instance. 
        Each key is the name of the serving plugin (AI engine). 
        Each value is a list where each item in the list is an inference.

        Example:
          {
            'object_detection_model' : [
              [{'TLBR_POS' : ...}, {'TLBR_POS' : ...}, {'TLBR_POS' : ...}],
              [{'TLBR_POS' : ...}]
            ],

            'anomaly_detection_model' : [
              'True/False'
            ]
          }
      """
      dct_inferences = self.dataapi_full_input().get('INFERENCES', {})
      if squeeze and len(dct_inferences) == 1:
        model_name = list(dct_inferences.keys())[0]
        result = dct_inferences[model_name]
      else:
        result = dct_inferences
      return result
    
        
    def dataapi_inferences_by_model(self, model_name : str) -> list:
      """
      Returns the inference results for a specific model.

      Parameters
      ----------
      model_name : str
        The name of the model for which the inference results are requested.

      Returns
      -------
      list
        The inference results.
      """
      dct_results = self.dataapi_inferences(squeeze=False)
      result = dct_results.get(model_name, [])
      return result
    
    
    def dataapi_inference_results(self, model_name : str = None, idx : int = 0) -> list:
      """
      Returns the inference results for a specific model and a specific input index.
      
      Parameters
      ----------
      model_name : str
        The name of the model for which the inference results are requested.
      
      idx : int, optional
        The index of the input for which the inference results are requested.
        The default value is 0.
      
      Returns
      -------
      list
        The inference results.
      """
      results = []
      raw_results = self.dataapi_inferences(squeeze=True)
      if model_name is not None and isinstance(raw_results, dict):
        results = raw_results.get(model_name, [])
      else:
        results = raw_results
        
      if len(results) > idx and isinstance(results, list):
        result = results[idx]
      else:
        result = results
      return result
    
    
    def dataapi_inference_result(self):
      return self.dataapi_inferences(squeeze=True)
    

    def dataapi_inferences_meta(self):
      """
      Returns
      -------
      dict{str:dict}
        the inference metadata that comes from the serving plugins configured for the current plugin instance

        Example:
          {
            'object_detection_model' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : 'IMG'},
            'anomaly_detection_model' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : 'STRUCT_DATA'}
          }
      """
      return self.dataapi_full_input().get('INFERENCES_META', {})

  """
  Section for methods that handle inputs / substreams.
  """
  if True:
    def dataapi_specific_input(self, idx=0, raise_if_error=False):
      """
      API for accessing a specific index (by its index in the 'INPUTS' list).

      Parameters
      ----------
      idx : int, optional
        The index of the input in the 'INPUTS' list
        The default value is 0.

      raise_if_error : bool, optional
        Whether to raise IndexError or not when the requested index is out of range.
        The default value is False

      Returns
      -------
      dict
        The requested input / substream.
        For the second example in the class docstring ('multi_modal_2_images_one_sensor_stream'), if `idx==0`, the API will return
          {
            'IMG' : np.ndarray(1),
            'STRUCT_DATA' : None,
            'INIT_DATA' : None,
            'TYPE' : 'IMG',
            'METADATA' : {Dictionary with current input metadata}
          }

      Raises
      ------
      IndexError
        when the requested index is out of range
      """
      lst_inputs = self.dataapi_inputs()
      if idx >= len(lst_inputs):
        if raise_if_error:
          raise IndexError("_DataAPIMixin - requested index {} in `dataapi_specific_input` is wrong".format(idx))
        else:
          return

      return lst_inputs[idx]

    def dataapi_images(self, full=False):
      """
      API for accessing all the images in the 'INPUTS' list
      Parameters
      ----------
      full : bool, optional
        Specifies whether the images are returned full (the whole input dictionary) or not (just the value of 'IMG' in the input dictionary)
        The default value is False

      Returns
      -------
      dict{int : dict} (if full==True) / dict{int : np.ndarray} (if full=False)
        The substreams that have images.
        For the second example in the class docstring ('multi_modal_2_images_one_sensor_stream'), the API will return
          {
            0 : {
              'IMG' : np.ndarray(1),
              'STRUCT_DATA' : None,
              'INIT_DATA' : None,
              'TYPE' : 'IMG',
              'METADATA' : {Dictionary with current input metadata}
            },

            1 : {
              'IMG' : np.ndarray(2),
              'STRUCT_DATA' : None,
              'INIT_DATA' : None,
              'TYPE' : 'IMG',
              'METADATA' : {Dictionary with current input metadata}
            }
          } if full==True

          or

          {
            0 : np.ndarray(1),
            1 : np.ndarray(2)
          } if full==False
      """
      lst_inp_img = list(filter(lambda x: x['TYPE'] == 'IMG', self.dataapi_inputs()))
      dct_idx_to_img = {i : x for i,x in enumerate(lst_inp_img)}

      if not full:
        dct_idx_to_img = {i : x['IMG'] for i,x in dct_idx_to_img.items()}

      return dct_idx_to_img
    
    def dataapi_images_as_list(self):
      lst_inp_imgs = [x['IMG'] for x in self.dataapi_inputs() if x['TYPE'] == 'IMG']
      return lst_inp_imgs

    def dataapi_specific_image(self, idx=0, full=False, raise_if_error=False):
      """
      API for accessing a specific image in the 'INPUTS' list

      Parameters
      ----------
      idx : int, optional
        The index of the image in the images list
        Attention! If there is a metastream that collects 3 inputs - ['IMG', 'STRUCT_DATA', 'IMG'], for accessing the last
        image, `idx` should be 1!
        The default value is 0

      full : bool, optional
        Passed to `dataapi_images`
        The default value is False

      raise_if_error : bool, optional
        Whether to raise IndexError or not when the requested index is out of range.
        The default value is False

      Returns
      -------
      dict (if full==True) / np.ndarray (if full==False)
        dict -> the whole input dictionary
        np.ndarray -> the value of 'IMG' in the input dictionary

      Raises
      ------
      IndexError
        when the requested index is out of range
      """
      dct_images = self.dataapi_images(full=full)
      if idx not in dct_images:
        if raise_if_error:
          raise IndexError("_DataAPIMixin - requested index {} in `dataapi_specific_image` is wrong".format(idx))
        else:
          return

      res = dct_images[idx]
      return res

    def dataapi_image(self, full=False, raise_if_error=False):
      """
      API for accessing the first image in the 'INPUTS' list
      (shortcut for `dataapi_specific_image`, most of the cases will have a single image on a stream)

      Parameters
      ----------
      full : bool, optional
        Passed to `dataapi_specific_image`
        The default value is False

      raise_if_error : bool, optional
        Passed to `dataapi_specific_image`
        The default value is False

      Returns
      -------
      dict (if full==True) / np.ndarray (if full==False)
        Returned by `dataapi_specific_image`

      Raises
      ------
      IndexError
        when there are no images on the stream if `raise_if_error == True`
      """
      return self.dataapi_specific_image(idx=0, full=full, raise_if_error=raise_if_error)

    def dataapi_struct_datas(self, full=False, as_list=False):
      """
      API for accessing all the structured datas in the 'INPUTS' list

      Parameters
      ----------
      full : bool, optional
        Specifies whether the structured datas are returned full (the whole input dictionary) or not (just the value of 'STRUCT_DATA' in the input dictionary)
        The default value is False
        
      as_list : bool, optional
        Specifies whether the structured datas are returned as a list or not - will work only if `full==False`

      Returns
      -------
      if as_list==False:
      
        dict{int : dict} (if full==True) / dict{int : object} (if full==False)
          The substreams that have structured data.
          For the second example in the class docstring ('multi_modal_2_images_one_sensor_stream'), the API will return
            {
              0: {
                'IMG' : None,
                'STRUCT_DATA' : an_object_or_most_likely_a_dict,
                'INIT_DATA' : None,
                'TYPE' : 'STRUCT_DATA',
                'METADATA' : {Dictionary with current input metadata}
              }
            } if full==True

            or

            {
              0 : an_object_or_most_likely_a_dict
            } if full==False
            
      if as_list==True:
      
        list[an_object_or_most_likely_a_dict] 
      
      
      Usage examples
      --------------
      
      ```
        datas = dataapi_struct_datas(full=True, as_list=False)
        indexes = list(datas.keys())
        for idx in indexes:
          struct_data = datas[idx]
          # do something with struct_data
          
          
        datas = dataapi_struct_datas(full=False, as_list=True)
        for struct_data in datas:
          # do something with struct_data
        
      """
      lst_inp_struct_data = list(filter(lambda x: x['TYPE'] == 'STRUCT_DATA', self.dataapi_inputs()))
      dct_idx_to_struct_data = {i : x for i,x in enumerate(lst_inp_struct_data)}

      if not full:
        dct_idx_to_struct_data = {i : x['STRUCT_DATA'] for i,x in dct_idx_to_struct_data.items()}

      keys = list(dct_idx_to_struct_data.keys())
      for k in keys:
        v = dct_idx_to_struct_data[k]
        if (hasattr(v, '__len__') and len(v) == 0) or (v is None):
          dct_idx_to_struct_data.pop(k)
      #endfor
      
      if as_list and not full:
        result = []
        for i in range(len(dct_idx_to_struct_data)):
          result.append(dct_idx_to_struct_data[i])
      else:
        result = dct_idx_to_struct_data

      return result


    def dataapi_specific_struct_data(self, idx=0, full=False, raise_if_error=False):
      """
      API for accessing a specific structured data in the 'INPUTS' list

      Parameters
      ----------
      idx : int, optional
        The index of the structured data in the structured datas list
        Attention! If there is a metastream that collects 3 inputs - ['IMG', 'STRUCT_DATA', 'IMG'], for accessing the structured data
        `idx` should be 0!
        The default value is 0

      full : bool, optional
        Passed to `dataapi_struct_datas`
        The default value is False

      raise_if_error : bool, optional
        Whether to raise IndexError or not when the requested index is out of range.
        The default value is True

      Returns
      -------
      dict (if full==True) / object (if full==False)
        dict -> the whole input dictionary
        object -> the value of 'STRUCT_DATA' in the input dictionary

      Raises
      ------
      IndexError
        when the requested index is out of range
      """
      dct_struct_datas = self.dataapi_struct_datas(full=full)
      if idx not in dct_struct_datas:
        if raise_if_error:
          raise IndexError("_DataAPIMixin - requested index {} in `dataapi_specific_struct_data` is wrong".format(idx))
        else:
          return

      res = dct_struct_datas[idx]
      return res

    def dataapi_struct_data(self, full=False, raise_if_error=False):
      """
      API for accessing the first structured data in the 'INPUTS' list
      (shortcut for `dataapi_specific_struct_data`, most of the cases will have a single structured data on a stream)

      Parameters
      ----------
      full : bool, optional
        Passed to `dataapi_specific_struct_data`
        The default value is False

      raise_if_error : bool, optional
        Passed to `dataapi_specific_struct_data`
        The default value is True

      Returns
      -------
      dict (if full==True) / object (if full==False)
        Returned by `dataapi_specific_struct_data`

      Raises
      ------
      IndexError
        when there are no struct_data on the stream
      """
      return self.dataapi_specific_struct_data(idx=0, full=full, raise_if_error=raise_if_error)

  """
  Section for methods that handle metadata
  """
  if True:
    def dataapi_inputs_metadata(self, as_list=False):
      """
      API for accessing the concatenated metadata from all inputs (images and structured datas together)
      This is not the same as the stream metadata that points to the overall params of the execution
      pipeline.

      Returns
      -------
      dict
        the concatenated metadata from all inputs
      """
      if as_list:
        metadata = []
        for inp in self.dataapi_inputs():
          metadata.append(inp['METADATA'])
      else:
        metadata = {}
        for inp in self.dataapi_inputs():
          metadata = {**metadata, **inp['METADATA']}
      return metadata

    def dataapi_specific_input_metadata(self, idx=0, raise_if_error=False):
      """
      API for accessing the metadata of a specific input

      Parameters
      ----------
      idx : int, optional
        Passed to `dataapi_specific_input`
        The default value is 0

      raise_if_error : bool, optional
        Passed to `dataapi_specific_input`
        The default value is False

      Returns
      -------
      dict
        the value of "METADATA" key in the requested input

      Raises
      ------
      IndexError
        when the requested index is out of range
      """
      inp = self.dataapi_specific_input(idx=idx, raise_if_error=raise_if_error)
      if inp is None:
        return
      return inp['METADATA']

    def dataapi_input_metadata(self, raise_if_error=False):
      """
      API for accessing the metadata of the first input
      (shortcut for `dataapi_specific_input_metadata`, most of the cases will have a single input on a stream)

      Parameters
      ----------
      raise_if_error : bool, optional
        Passed to `dataapi_specific_input_metadata`
        The default value is False

      Returns
      -------
      dict
        Returned by `dataapi_specific_input_metadata`

      Raises
      ------
      IndexError
        raised by `dataapi_specific_input_metadata`
      """
      return self.dataapi_specific_input_metadata(idx=0, raise_if_error=raise_if_error)

    def dataapi_all_metadata(self):
      """
      API for accessing the concatenated stream metadata and metadata from all inputs

      Returns
      -------
      dict
        the concatenated stream metadata and metadata from all inputs
      """
      metadata = {
        **self.dataapi_stream_metadata(),
        **self.dataapi_inputs_metadata()
      }
      return metadata

  """
  Section for methods that handle initial data
  """
  if True:
    def dataapi_specific_input_init_data(self, idx=0, raise_if_error=False):
      """
      API for accessing the initial data of a specific input

      Parameters
      ----------
      idx : int, optional
        Passed to `dataapi_specific_input`
        The default value is 0

      raise_if_error : bool, optional
        Passed to `dataapi_specific_input`
        The default value is False

      Returns
      -------
      dict
        the value of "INIT_DATA" key in the requested input

      Raises
      ------
      IndexError
        when the requested index is out of range
      """
      inp = self.dataapi_specific_input(idx=idx, raise_if_error=raise_if_error)
      if inp is None:
        return
      return inp['INIT_DATA']

  """
  Section for methods that handle images inferences
  """
  if True:
    def dataapi_images_inferences(self):
      """
      API for accessing just the images inferences.
      Filters the output of `dataapi_inferences`, keeping only the AI engines that run on images

      Returns
      -------
      dict{str:list}
        the inferences that comes from the images serving plugins configured for the current plugin instance.
      """
      dct_inferences = self.dataapi_inferences()
      dct_inferences_meta = self.dataapi_inferences_meta()

      filtered_dct_inferences = {}
      for k, v in dct_inferences.items():
        if dct_inferences_meta[k]['PICKED_INPUT'] == 'IMG':
          filtered_dct_inferences[k] = v

      return filtered_dct_inferences

    def dataapi_images_global_inferences(self):
      """
      Alias for `dataapi_images_inferences`
      """
      return self.dataapi_images_inferences()

    def dataapi_images_plugin_inferences(self):
      """
      API for accessing the images inferences, filtered by confidence threshold and object types.
      More specifically, all the plugin inferences are the global inferences that surpass a configured confidence
      threshold and have a specific type. For example, an object detector basically infers for all the objects in
      COCO dataset. But, a certain plugin may need only persons and dogs.

      Returns
      -------
      dict{str:list}
        filtered images inferences by confidence threshold and object types
      """
      return self._pre_process_outputs['DCT_PLUGIN_INFERENCES']

    def dataapi_images_instance_inferences(self):
      """
      API for accessing the images inferences, filtered by confidence threshold, object types and target zone.
      More specifically, all the instance inferences are the plugin inferences that intersects (based on PRC_INTERSECT)
      with the configured target zone.

      Returns
      -------
      dict{str:list}
        filtered images inferences by confidence threshold, object types and target zone
      """
      return self._pre_process_outputs['DCT_INSTANCE_INFERENCES']

    def dataapi_images_plugin_positional_inferences(self):
      """
      API for accessing the images inferences that have positions (TLBR_POS).
      Returns
      -------
      dict{str:list}
        filtered images inferences by having positions (TLBR_POS)
      """
      return self._pre_process_outputs['DCT_PLUGIN_POSITIONAL_INFERENCES']

    def dataapi_specific_image_inferences(self, idx=0, how=None, mode=None, raise_if_error=False):
      """
      API for accesing inferences for a specific image (global, plugin or instance inferences)
      See `dataapi_images_global_inferences`, `dataapi_images_plugin_inferences`, `dataapi_images_instance_inferences`

      Parameters
      ----------
      idx : int, optional
        The index of the image in the images list
        Attention! If there is a metastream that collects 3 inputs - ['IMG', 'STRUCT_DATA', 'IMG'], for accessing the last
        image, `idx` should be 1!
        The default value is 0

      how : str, optional
        Could be: 'list' or 'dict'
        Specifies how the inferences are returned. If 'list', then the AI engine information will be lost and all the
        inferences from all the employed AI engines will be concatenated in a list; If 'dict', then the AI engine information
        will be preserved.
        The default value is None ('list')

      mode : str, optional
        Could be: 'global', 'plugin' or 'instance'
        Specifies which inferences are requested.
        The default value is None ('instance')

      raise_if_error : bool, optional
        Whether to raise IndexError or not when the requested index is out of range.
        The default value is False

      Returns
      -------
      dict (if how == 'dict') or list (if how == 'list')
        the requested image inferences (global, plugin or instance) in the requested format (dict or list)

      Raises
      ------
      IndexError
        when the requested index is out of range
      """
      if how is None:
        how = 'list'

      if mode is None:
        mode = 'instance'

      assert how in ['dict', 'list']
      assert mode in ['global', 'plugin', 'instance', 'plugin_positional']
      img = self.dataapi_specific_image(idx=idx, raise_if_error=raise_if_error)
      if img is None:
        return

      dct_inferences = None
      if mode == 'global':
        dct_inferences = self.dataapi_images_global_inferences()
      elif mode == 'plugin':
        dct_inferences = self.dataapi_images_plugin_inferences()
      elif mode == 'instance':
        dct_inferences = self.dataapi_images_instance_inferences()
      elif mode == 'plugin_positional':
        dct_inferences = self.dataapi_images_plugin_positional_inferences()
      #endif

      dct_inferences_idx = {
        model : lst_2d[idx] for model, lst_2d in dct_inferences.items()
      }

      if how == 'dict':
        return dct_inferences_idx
      elif how == 'list':
        return self.log.flatten_2d_list(list(dct_inferences_idx.values()))
      #endif

    def dataapi_image_inferences(self, how=None, mode=None, raise_if_error=False):
      """
      API for accessing the first image inferences
      (shortcut for `dataapi_specific_image_inferences`, most of the cases will have a single image on a stream)

      Parameters
      ----------
      how : str, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is None

      mode : str, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is None

      raise_if_error : bool, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is False

      Returns
      -------
      dict (if how == 'dict') or list (if how == 'list')
        returned by `dataapi_specific_image_inferences`

      Raises
      ------
      IndexError
        raised by `dataapi_specific_image_inferences`
      """
      return self.dataapi_specific_image_inferences(idx=0, how=how, mode=mode, raise_if_error=raise_if_error)

    def dataapi_specific_image_global_inferences(self, idx=0, how=None, raise_if_error=False):
      """
      API for accessing a specific image global inferences
      (shortcut for `dataapi_specific_image_inferences`)

      Parameters
      ----------
      idx : int, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is None

      how : str, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is None

      raise_if_error : bool, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is False

      Returns
      -------
      dict (if how == 'dict') or list (if how == 'list')
        returned by `dataapi_specific_image_inferences`

      Raises
      ------
      IndexError
        raised by `dataapi_specific_image_inferences`
      """
      return self.dataapi_specific_image_inferences(idx=idx, how=how, mode='global', raise_if_error=raise_if_error)

    def dataapi_specific_image_plugin_inferences(self, idx=0, how=None, raise_if_error=False):
      """
      API for accessing a specific image plugin inferences
      (shortcut for `dataapi_specific_image_inferences`)

      Parameters
      ----------
      idx : int, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is None

      how : str, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is None

      raise_if_error : bool, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is False

      Returns
      -------
      dict (if how == 'dict') or list (if how == 'list')
        returned by `dataapi_specific_image_inferences`

      Raises
      ------
      IndexError
        raised by `dataapi_specific_image_inferences`
      """
      return self.dataapi_specific_image_inferences(idx=idx, how=how, mode='plugin', raise_if_error=raise_if_error)

    def dataapi_specific_image_instance_inferences(self, idx=0, how=None, raise_if_error=False):
      """
      API for accessing a specific image inferences for the current plugin instance
      (shortcut for `dataapi_specific_image_inferences`)

      Parameters
      ----------
      idx : int, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is None

      how : str, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is None ('list')

      raise_if_error : bool, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is False

      Returns
      -------
      dict (if how == 'dict') or list (if how == 'list')
        returned by `dataapi_specific_image_inferences`

      Raises
      ------
      IndexError
        raised by `dataapi_specific_image_inferences`
      """
      return self.dataapi_specific_image_inferences(idx=idx, how=how, mode='instance', raise_if_error=raise_if_error)

    def dataapi_specific_image_plugin_positional_inferences(self, idx=0, how=None, raise_if_error=False):
      """
      API for accessing a specific image plugin positional inferences
      (shortcut for `dataapi_specific_image_inferences`)

      Parameters
      ----------
      idx : int, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is None

      how : str, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is None

      raise_if_error : bool, optional
        Passed to `dataapi_specific_image_inferences`
        The default value is False

      Returns
      -------
      dict (if how == 'dict') or list (if how == 'list')
        returned by `dataapi_specific_image_inferences`

      Raises
      ------
      IndexError
        raised by `dataapi_specific_image_inferences`
      """
      return self.dataapi_specific_image_inferences(idx=idx, how=how, mode='plugin_positional', raise_if_error=raise_if_error)

    def dataapi_image_global_inferences(self, how=None, raise_if_error=False):
      """
      API for accessing the first image global inferences
      (shortcut for `dataapi_specific_image_global_inferences`, most of the cases will have a single image on a stream)

      Parameters
      ----------
      how : str, optional
        Passed to `dataapi_specific_image_global_inferences`
        The default value is None

      raise_if_error : bool, optional
        Passed to `dataapi_specific_image_global_inferences`
        The default value is False

      Returns
      -------
      dict (if how == 'dict') or list (if how == 'list')
        returned by `dataapi_specific_image_global_inferences`

      Raises
      ------
      IndexError
        raised by `dataapi_specific_image_global_inferences`
      """
      return self.dataapi_specific_image_global_inferences(idx=0, how=how, raise_if_error=raise_if_error)

    def dataapi_image_plugin_inferences(self, how=None, raise_if_error=False):
      """
      API for accessing the first image plugin inferences
      (shortcut for `dataapi_specific_image_plugin_inferences`, most of the cases will have a single image on a stream)

      Parameters
      ----------
      how : str, optional
        Passed to `dataapi_specific_image_plugin_inferences`
        The default value is None

      raise_if_error : bool, optional
        Passed to `dataapi_specific_image_plugin_inferences`
        The default value is False

      Returns
      -------
      dict (if how == 'dict') or list (if how == 'list')
        returned by `dataapi_specific_image_plugin_inferences`

      Raises
      ------
      IndexError
        raised by `dataapi_specific_image_plugin_inferences`
      """
      return self.dataapi_specific_image_plugin_inferences(idx=0, how=how, raise_if_error=raise_if_error)

    def dataapi_image_instance_inferences(self, how=None, raise_if_error=False):
      """
      API for accessing the first image instance inferences
      (shortcut for `dataapi_specific_image_instance_inferences`, most of the cases will have a single image on a stream)

      Parameters
      ----------
      how : str, optional
        Passed to `dataapi_specific_image_instance_inferences`
        The default value is None ('list')

      raise_if_error : bool, optional
        Passed to `dataapi_specific_image_instance_inferences`
        The default value is False

      Returns
      -------
      dict (if how == 'dict') or list (if how == 'list')
        returned by `dataapi_specific_image_instance_inferences`

      Raises
      ------
      IndexError
        raised by `dataapi_specific_image_instance_inferences`
      """
      return self.dataapi_specific_image_instance_inferences(idx=0, how=how, raise_if_error=raise_if_error)

    def dataapi_image_plugin_positional_inferences(self, how=None, raise_if_error=False):
      """
      API for accessing the first image plugin positional inferences
      (shortcut for `dataapi_specific_image_plugin_positional_inferences`, most of the cases will have a single image on a stream)

      Parameters
      ----------
      how : str, optional
        Passed to `dataapi_specific_image_plugin_positional_inferences`
        The default value is None

      raise_if_error : bool, optional
        Passed to `dataapi_specific_image_plugin_positional_inferences`
        The default value is False

      Returns
      -------
      dict (if how == 'dict') or list (if how == 'list')
        returned by `dataapi_specific_image_plugin_positional_inferences`

      Raises
      ------
      IndexError
        raised by `dataapi_specific_image_plugin_positional_inferences`
      """
      return self.dataapi_specific_image_plugin_positional_inferences(idx=0, how=how, raise_if_error=raise_if_error)

  """
  Section for methods that handle struct datas inferences
  """
  if True:
    def dataapi_struct_datas_inferences(self):
      """
      API for accessing just the structured datas inferences.
      Filters the output of `dataapi_inferences`, keeping only the AI engines that run on structured datas

      Returns
      -------
      dict{str:list}
        the inferences that comes from the structured datas serving plugins configured for the current plugin instance.
      """
      dct_inferences = self.dataapi_inferences()
      dct_inferences_meta = self.dataapi_inferences_meta()

      filtered_dct_inferences = {}
      for k, v in dct_inferences.items():
        if dct_inferences_meta[k]['PICKED_INPUT'] == 'STRUCT_DATA':
          filtered_dct_inferences[k] = v

      return filtered_dct_inferences

    def dataapi_specific_struct_data_inferences(self, idx=0, how=None, raise_if_error=False):
      """
      API for accesing a specific structured data inferences

      Parameters
      ----------
      idx : int, optional
        The index of the structured data in its list
        Attention! If there is a metastream that collects 3 inputs - ['IMG', 'STRUCT_DATA', 'IMG'], for accessing the structured data,
        `idx` should be 0!
        The default value is 0

      how : str, optional
        Could be: 'list' or 'dict'
        Specifies how the inferences are returned. If 'list', then the AI engine information will be lost and all the
        inferences from all the employed AI engines will be concatenated in a list; If 'dict', then the AI engine information
        will be preserved.
        The default value is None ('list')

      raise_if_error : bool, optional
        Whether to raise IndexError or not when the requested index is out of range.
        The default value is False

      Returns
      -------
      dict (if how == 'dict') or list (if how == 'list')
        the requested structured data inferences in the requested format (dict or list)

      Raises
      ------
      IndexError
        when the requested index is out of range
      """
      if how is None:
        how = 'list'

      assert how in ['dict', 'list']

      sd = self.dataapi_specific_struct_data(idx=idx, raise_if_error=raise_if_error)
      if sd is None:
        return

      dct_inferences = self.dataapi_struct_datas_inferences()
      dct_inferences_idx = {
        model : lst[idx] for model, lst in dct_inferences.items()
      }

      if how == 'dict':
        return dct_inferences_idx
      elif how == 'list':
        return list(dct_inferences_idx.values())
      #endif

    def dataapi_struct_data_inferences(self, how=None, raise_if_error=False):
      """
      API for accesing a the first structured data inferences
      (shortcut for `dataapi_specific_struct_data_inferences`, most of the cases will have a single struct data on a stream)

      Parameters
      ----------
      how : str, optional
        Passed to `dataapi_specific_struct_data_inferences`
        The default value is None

      raise_if_error : bool, optional
        Passed to `dataapi_specific_struct_data_inferences`
        The default value is False

      Returns
      -------
      dict (if how == 'dict') or list (if how == 'list')
        returned by `dataapi_specific_struct_data_inferences`

      Raises
      ------
      IndexError
        raised by `dataapi_specific_struct_data_inferences`
      """
      return self.dataapi_specific_struct_data_inferences(idx=0, how=how, raise_if_error=raise_if_error)


import os
import pandas as pd
import numpy as np

from time import time
from datetime import timedelta

from naeural_core import constants as ct
from naeural_core.utils.datetime_utils import add_microseconds_to_str_timedelta

class _LimitedDataMixin(object):

  def __init__(self):
    self._limited_data_finished = False    
    self.last_limited_data_counter = -1
    self.last_limited_data_total_counter = -1
    self.last_limited_data_progress = -1
    self.last_limited_data_remaining_time = -1 
    self.last_limited_data_finished_flag = None
    self.is_data_limited = False
    
    self._payload_history = [] # list of payloads for limited streams
    super(_LimitedDataMixin, self).__init__()
    return

  @property
  def is_limited_data_finished(self):
    # TODO: `_limited_data_finished` is set to `True` by `cv` -> move to `base`!
    return self._limited_data_finished

  @property
  def is_data_limited_and_has_frame(self):
    metadata = self.dataapi_all_metadata()
    if metadata is not None:
      total = metadata.get('frame_count', 0)
      if total is not None and total > 0:
        self.is_data_limited = True
        return True
    return False

  @property
  def is_last_data(self):
    if self.is_data_limited_and_has_frame:
      curr = self.limited_data_frame_current
      total = self.limited_data_frame_count
      if curr >= total:
        return True
    return False
  
  @property
  def was_last_data(self):
    if self.last_limited_data_counter > -1:
      curr = self.last_limited_data_counter
      total = self.last_limited_data_total_counter
      if curr >= total:
        return True
    return False  
  
  @property
  def limited_data_finished_flag(self):
    if self.is_data_limited_and_has_frame:
      self.last_limited_data_finished_flag = self.dataapi_all_metadata()['finished']
      return self.last_limited_data_finished_flag
    return None

  @property
  def limited_data_counter(self):
    if self.is_data_limited_and_has_frame:
      self.last_limited_data_counter = self.dataapi_all_metadata()['frame_current']
      return self.last_limited_data_counter
    return

  @property
  def limited_data_frame_current(self):
    return self.limited_data_counter

  @property
  def limited_data_total_counter(self):
    if self.is_data_limited_and_has_frame:
      self.last_limited_data_total_counter = self.dataapi_all_metadata()['frame_count']
      return self.last_limited_data_total_counter
    return

  @property
  def limited_data_frame_count(self):
    return self.limited_data_total_counter

  @property
  def limited_data_progress(self):
    if self.limited_data_counter is not None:
      crt_frame = self.limited_data_counter
      total_frames = self.limited_data_total_counter
      progress = crt_frame / total_frames * 100
      progress = round(progress, 3)
      self.last_limited_data_progress = progress
      return progress
    return -1

  @property
  def limited_data_remaining_time(self):
    res = None
    if self.is_data_limited_and_has_frame:
      diff_time = time() - self.first_process_time
      crt_frame = self.limited_data_counter
      total_frames = self.limited_data_total_counter
      remaining_time = (total_frames - crt_frame) * (diff_time / crt_frame)
      self.last_limited_data_remaining_time = remaining_time
      res = remaining_time
    elif self.last_limited_data_remaining_time > -1:
      res = self.last_limited_data_remaining_time
    return res

  @property
  def limited_data_fps(self):
    res = None
    all_metadata = self.dataapi_all_metadata()    
    if self.is_data_limited_and_has_frame:
      try:
        res = all_metadata['cap_resolution']
      except KeyError:
        res = all_metadata['fps']
    return res

  @property
  def limited_data_seconds_elapsed(self):
    res = None
    if self.is_data_limited_and_has_frame:
      res = np.array(self.limited_data_frame_current) / np.array(self.limited_data_fps) # scalars, but safer with numpy
    return res

  @property
  def limited_data_crt_time(self):
    res = None
    if self.is_data_limited_and_has_frame:
      crt_time = add_microseconds_to_str_timedelta(str(timedelta(seconds=self.limited_data_seconds_elapsed)))
      res = crt_time
    return res

  @property
  def limited_data_duration(self):
    res = None
    if self.is_data_limited_and_has_frame:
      res = np.array(self.limited_data_total_counter) / np.array(self.limited_data_fps) # scalars, but safer with numpy
    return res
  
  @property
  def limited_data_process_fps(self):
    res = 0
    if self.is_data_limited_and_has_frame:
      try:
        elapsed_time = time() - self.first_process_time
        crt_frame = self.limited_data_counter
        res = round(crt_frame / elapsed_time, 2)
      except:
        pass
    return res

  def _patch_limited_data_history(self, payload, debug_objects):
    """
    Patch created especially because one of our implementer did not want to track objects in a finite stream
    """
    if not self.is_data_limited_and_has_frame:
      return payload

    curr_count = self.limited_data_counter
    fps = self.dataapi_all_metadata()['fps'] if self.dataapi_all_metadata()['fps'] is not None else 0
    time_sec = curr_count / fps if fps != 0 else 0
    s_time = str(timedelta(seconds=int(time_sec)))
    # minimal load for each processing step:
    dct_payload = {
      ct.FRAME_CURRENT: curr_count,
      ct.STREAM_TIME: s_time,
      ct.DEBUG_OBJECTS: debug_objects,
      ct.ALERT_HELPER: self.get_alerter_status(),

      ct.DEBUG_OBJECTS_SUMMARY: self.get_debug_objects_summary(debug_objects),
      ct.CONFIDENCE_THRESHOLD: self._get_confidence_threshold(),
    }
    if payload is not None:
      dct_payload[ct.HAS_PAYLOAD] = True
      # add actual payload info in any
      for k, v in vars(payload).items():
        if k not in [ct.IMG, ]:  # maybe other large and not necesary data
          dct_payload[k] = v
    else:
      dct_payload[ct.HAS_PAYLOAD] = False
    self._payload_history.append(dct_payload)
    if self.is_last_data:
      # now we send the results!
      # Note: even if the stream is LIVE eventually the stream plugin will
      #       push the last data in stream in buffer and will stop after the
      #       capture manager will consume it so we ALWAYS receive the last
      #       observation. This way `is_last_data` is safe.
      if payload is None:
        payload = self._create_payload()

      if self._instance_config.get(ct.PROCESSING_RESULTS_CSV, True):
        file_path = os.path.join(
          self.log.get_output_folder(),
          '{}.csv'.format(self.log.now_str())
        )
        pd.DataFrame(self._payload_history).to_csv(file_path)

        # save to dropbox
        _results = {'UPLOAD_FILE' : file_path}
        if _results is None:
          _results = 'Error while uploading csv results'
      else:
        _results = self._payload_history
      vars(payload)[ct.HISTORY] = _results
      self._payload_history = []

    return payload



import abc
import json
from collections import deque

from naeural_core.data.base.base_plugin_dct import DataCaptureThread
from ratio1 import Session

_CONFIG = {
  **DataCaptureThread.CONFIG,

  "URL": None,
  'RECONNECTABLE': True,


  'CAP_RESOLUTION' : 1,
  'STREAM_WINDOW' : 1_000,
  'LIVE_FEED' : False,
  'FILTER_WORKERS': None,

  'VALIDATION_RULES': {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],
  },
}


class BaseDecentraiConnectorDataCapture(DataCaptureThread):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(BaseDecentraiConnectorDataCapture, self).__init__(**kwargs)
    self.message_queue = deque(maxlen=1000)
    self._session = None
    return

  def _init(self):
    # create a session object and connect to ut
    self._session = Session(
      host=None,
      port=None,
      user=None,
      pwd=None,
      name=f'{self._device_id}_{self.cfg_name}',
      config=self.shmem['config_communication']['PARAMS'],
      filter_workers=self.cfg_filter_workers,
      log=self.log,
      on_heartbeat=self.heartbeat_callback,
      on_notification=self.notification_callback,
      on_payload=self.payload_callback,
      bc_engine=self.shmem[self.ct.BLOCKCHAIN_MANAGER],
    )
    return

  def heartbeat_callback(self, current_session, e2id, payload):
    self.message_queue.append({
      "TYPE": "heartbeat",
      "E2ID": e2id,
      "ADDRESS": payload["EE_SENDER"],
      "DATA": dict(payload),
    })

  def notification_callback(self, current_session, e2id, payload):
    self.message_queue.append({
      "TYPE": "notification",
      "E2ID": e2id,
      "ADDRESS": payload["EE_SENDER"],
      "DATA": dict(payload),
    })

  def payload_callback(self, current_session, e2id, session, signature, instance_id, payload):
    self.message_queue.append({
      "TYPE": "payload",
      "E2ID": e2id,
      "ADDRESS": payload["EE_SENDER"],
      # "SESSION": session,
      # "SIGNATURE": signature,
      # "INSTANCE_ID": instance_id,
      "DATA": dict(payload),
    })

  def _release(self):
    self._session.close(wait_close=True)
    return

  def _maybe_reconnect(self):
    return

  def _run_data_aquisition_step(self):
    if len(self.message_queue) == 0:
      return

    no_messages = len(self.message_queue)
    messages = []
    for _ in range(no_messages):
      messages.append(self.message_queue.popleft())
    # endfor

    lst_dct_output = []

    for msg in messages:
      dct_output = self.interpret_message(msg)
      if dct_output is not None:
        lst_dct_output.append(dct_output)
    # endfor

    if len(lst_dct_output) == 0:
      return

    self._add_struct_data_input(lst_dct_output)
    return

  @abc.abstractmethod
  def interpret_message(self, message) -> dict:
    """
    Interpret a message received from the network. The message is structured like this:
    ```
    {
      "TYPE": one of the following ["payload", "notification", "heartbeat"],
      "E2ID": the node name,
      "DATA": the message received,
    }
    ```

    To ignore the message, return `None`

    Parameters
    ----------
    message : dict
        the message received from the network

    Returns
    -------
    interpreted_message(optional) : dict | None
        a structured data point with information.
    """
    raise NotImplementedError
